import io
from mb_capcha_ocr import OcrModel
from PIL import Image


class CapchaProcessing:
    """
    Base class for capcha processing for self implemented
    Examples:
    ```py
    class MyCapchaProcessing(CapchaProcessing):
        def process_image(self, img: bytes) -> str:
            return "my_text"
    ```
    """

    def __init__(self):
        return

    def process_image(self, img: bytes) -> str:
        """
        Process image and return text

        Args:
            img (bytes): image input as bytes

        Returns:
            success (str): text from image
        """
        raise NotImplementedError("process_image is not implemented")


class CapchaOCR(CapchaProcessing):
    """
    Onnx based OCR for capcha processing
    https://pypi.org/project/mb-capcha-ocr/

    Args:
        model_path (str, optional): path to model file
    """

    def __init__(self, model_path: str = None):
        super().__init__()
        # loading model will take about 1-3 seconds
        self.model = OcrModel(model_path)

    def process_image(self, img: bytes) -> str:
        """
        Process image and return text

        Args:
            img (bytes): image input as bytes

        Returns:
            success (str): text from image
        """
        image = Image.open(io.BytesIO(img))
        text = self.model.predict(image)
        return text





from mbbank import MBBank
import datetime

mb = MBBank(username="0943525578", password="Krisktan5436*****")
end_query_day = datetime.datetime.now()
start_query_day = end_query_day - datetime.timedelta(days=30)
# print(mb.userinfo())
print(mb.inquiryAccountName(creditAccount="9961384366", creditAccountType="ACCOUNT", bankCode="970436"))


import asyncio
import json
import math
import os
import time
import traceback
import wasmtime
from .helper import Memory, fs_object, process_object, dict_warper, hash_list
from contextvars import ContextVar

undefined = ContextVar('undefined')


class globalThis:
    def __init__(self):
        self.exports = undefined
        self.window = dict_warper({"document": dict_warper({})})
        self.fs = fs_object()
        self.process = process_object()
        self.location = dict_warper({"origin": "https://online.mbbank.com.vn"})

    def __getattribute__(self, item):
        if item == "Object":
            return object
        elif item == "Array":
            return list
        elif item == "Uint8Array":
            return bytes
        return object.__getattribute__(self, item)


global_this = globalThis()


class GO:
    def __init__(self, wasm_store):
        self.wasm_store = wasm_store
        self.argv = ["js"]
        self.env = {}
        self._exitPromise = asyncio.Event()
        self._pendingEvent = None
        self._scheduledTimeouts = {}
        self._nextCallbackTimeoutID = 1
        self.go_js = GOJS(self)

        def setInt64(addr, v):
            self.mem.setUint32(addr + 0, v, True)
            self.mem.setUint32(addr + 4, math.floor(v / 4294967296), True)

        self.setInt64 = setInt64

        def getInt64(addr):
            low = self.mem.getUint32(addr + 0, True)
            high = self.mem.getInt32(addr + 4, True)
            return low + high * 4294967296

        self.getInt64 = getInt64

        def loadValue(addr):
            f = self.mem.getFloat64(addr, True)
            if f == 0:
                return
            elif not math.isnan(f):
                return f
            aid = self.mem.getInt32(addr, True)
            return self._values[aid]

        self.loadValue = loadValue

        def storeValue(addr, v=undefined):
            nanHead = 0x7FF80000
            if type(v) in [int, float] and v != 0:
                if math.isnan(v):
                    self.mem.setInt32(addr + 4, nanHead, True)
                    self.mem.setFloat64(addr, 0, True)
                    return
                self.mem.setFloat64(addr, v, True)
                return

            if v is undefined:
                self.mem.setFloat64(addr, 0, True)
                return

            obj_id = self._ids.get(v, undefined)

            if obj_id == undefined:
                if self._idPool:
                    obj_id = self._idPool.pop()
                else:
                    obj_id = len(self._values)
                if obj_id >= len(self._values):
                    self._values.extend([undefined for _ in range(obj_id - len(self._values) + 1)])
                if obj_id >= len(self._goRefCounts):
                    self._goRefCounts.extend([float("inf") for _ in range(obj_id - len(self._goRefCounts) + 1)])
                self._values[obj_id] = v
                self._goRefCounts[obj_id] = 0
                self._ids.setdefault(v, obj_id)

            self._goRefCounts[obj_id] += 1
            typeFlag = 1  # everything is an object
            if v is None:
                typeFlag = 0
            elif type(v) is str:
                typeFlag = 2
            # elif type(v) is symbol: # symbol is not supported in python
            #    typeFlag = 3
            elif callable(v):
                typeFlag = 4
            self.mem.setInt32(addr + 4, nanHead | typeFlag, True)
            self.mem.setInt32(addr, obj_id, True)

        self.storeValue = storeValue

        def loadSlice(addr):
            array = getInt64(addr + 0)
            len_read = getInt64(addr + 8)
            return self.mem.read(array, array + len_read), array, len_read

        self.loadSlice = loadSlice

        def loadSliceOfValues(addr):
            array = getInt64(addr + 0)
            len_read = getInt64(addr + 8)
            return [self.loadValue(array + i * 8) for i in range(len_read)]

        self.loadSliceOfValues = loadSliceOfValues

        def loadString(addr):
            array = getInt64(addr + 0)
            len_read = getInt64(addr + 8)
            return self.mem.read(array, array + len_read).decode("utf-8")

        self.loadString = loadString

    @classmethod
    def exit_process(cls, exitCode):
        print("exit code:", exitCode)

    def importObject(self, imports_type: list[wasmtime.ImportType]):
        def proxy(name):
            def fn(*args, **kwargs):
                call = getattr(self.go_js, name)
                return call(*args, **kwargs)

            return fn

        return [wasmtime.Func(self.wasm_store, i.type, proxy(i.name)) for i in imports_type]

    # noinspection PyAttributeOutsideInit
    def run(self, inst):
        self._inst = inst
        self.mem = Memory(self.wasm_store, inst["mem"])
        self._values = [
            float("nan"),
            0,
            None,
            True,
            False,
            global_this,
            self
        ]
        self._goRefCounts = [float("inf") for _ in range(50)]
        self._ids = dict(
            [
                (0, 1),
                (None, 2),
                (True, 3),
                (False, 4),
                (global_this, 5),
                (self, 6),
            ]
        )
        self._idPool = []
        self.exited = False
        self.offset = 4096

        def strPtr(str_data: str):
            ptr = self.offset
            bytes_data = (str_data + "\0").encode()
            self.mem.write(bytes_data, self.offset)
            self.offset += len(bytes_data)
            if (self.offset % 8) != 0:
                self.offset += 8 - (self.offset % 8)
            return ptr

        argc = len(self.argv)
        argvPtrs = []
        [argvPtrs.append(strPtr(i)) for i in self.argv]
        argvPtrs.append(0)
        envKeys = sorted(self.env.keys(), key=str)
        [argvPtrs.append(strPtr(f"{i}={self.env[i]}")) for i in envKeys]
        argvPtrs.append(0)
        argv = self.offset
        for ptr in argvPtrs:
            self.mem.setUint32(self.offset, ptr, True)
            self.mem.setUint32(self.offset + 4, 0, True)
            self.offset += 8
        wasmMinDataAddr = 4096 + 8192
        if self.offset >= wasmMinDataAddr:
            raise MemoryError("total length of command line and environment variables exceeds limit")
        self.go_js.run_init()
        self._inst["run"](self.wasm_store, argc, argv)
        if self.exited:
            self._exitPromise.set()

    def _resume(self):
        if self.exited:
            self._exitPromise.set()
            raise RuntimeError("Go program has already exited")
        self._inst["resume"](self.wasm_store)
        if self.exited:
            self._exitPromise.set()

    def _makeFuncWrapper(self, ids):
        def wrapper(*args, **kwargs):
            event = dict_warper({"id": int(ids), "args": hash_list(args), "this": global_this})
            self._pendingEvent = event
            self._resume()
            return event.result

        return wrapper


# noinspection PyAttributeOutsideInit
class GOJS:
    def __init__(self, go_obj: GO):
        self.go = go_obj

    def run_init(self):
        self.mem = self.go.mem

    def __getattribute__(self, item):
        if item.startswith("runtime."):
            item = item.replace("runtime.", "rt_")
        elif item.startswith("syscall/js."):
            item = item.replace("syscall/js.", "sysjs_")
        data = object.__getattribute__(self, item)
        return data

    def rt_wasmExit(self, sp):
        code = self.mem.getInt32(sp + 8, True)
        self.go.exited = True
        self.go._exitPromise.set()
        self.go.exit_process(code)

    def rt_wasmWrite(self, sp):
        sp = (sp >> 0) & 0xFFFFFFFF
        fd = self.go.getInt64(sp + 8)
        p = self.go.getInt64(sp + 16)
        n = self.mem.getInt32(sp + 24, True)
        global_this.fs.writeSync(fd, self.mem.read(p, p + n))

    def rt_resetMemoryDataView(self, address):
        # do nothing maybe ?
        pass

    def rt_nanotime1(self, sp):
        sp = (sp >> 0) & 0xFFFFFFFF
        self.go.setInt64(sp + 8, int(time.time() * 1000) * 1000000)

    def rt_walltime(self, sp):
        sp = (sp >> 0) & 0xFFFFFFFF
        msec = int(time.time() * 1000)
        self.go.setInt64(sp + 8, msec / 1000)
        self.mem.setInt32(sp + 16, (msec % 1000) * 1000000, True)

    def rt_scheduleTimeoutEvent(self, *args):
        pass

    def rt_clearTimeoutEvent(self, *args):
        pass

    def rt_getRandomData(self, sp):
        sp = (sp >> 0) & 0xFFFFFFFF
        byte_aray, start, length = self.go.loadSlice(sp + 8)
        random_data = os.urandom(length)
        self.mem.write(random_data, start)

    def sysjs_finalizeRef(self, sp):
        sp = (sp >> 0) & 0xFFFFFFFF
        ids = self.mem.getUint32(sp + 8, True)
        if self.go._goRefCounts[ids] == 0:
            v = self.go._values[ids]
            self._values[ids] = None
            self.go._ids.pop(v, None)
            self.go._idPool.append(ids)

    def sysjs_stringVal(self, sp):
        sp = (sp >> 0) & 0xFFFFFFFF
        data = self.go.loadString(sp + 8)
        self.go.storeValue(sp + 24, data)

    def sysjs_valueGet(self, sp):
        sp = (sp >> 0) & 0xFFFFFFFF
        obj = self.go.loadValue(sp + 8)
        name = self.go.loadString(sp + 16)
        if type(obj) is dict:
            result = obj.get(name, undefined)
        else:
            result = getattr(obj, name)
        sp = self.go._inst["getsp"](self.go.wasm_store) >> 0
        self.go.storeValue(sp + 32, result)

    def sysjs_valueSet(self, sp):
        sp = (sp >> 0) & 0xFFFFFFFF
        setattr(self.go.loadValue(sp + 8), self.go.loadString(sp + 16), self.go.loadValue(sp + 32))

    def sysjs_valueDelete(self, sp):
        sp = (sp >> 0) & 0xFFFFFFFF
        delattr(self.go.loadValue(sp + 8), self.go.loadString(sp + 16))

    def sysjs_valueIndex(self, sp):
        sp = (sp >> 0) & 0xFFFFFFFF
        obj = self.go.loadValue(sp + 8)
        int_data = self.go.getInt64(sp + 16)
        result = obj.get(int_data, undefined)
        self.go.storeValue(sp + 24, result)

    def sysjs_valueSetIndex(self, *args):
        pass

    def sysjs_valueCall(self, sp):
        sp = (sp >> 0) & 0xFFFFFFFF
        try:
            v = self.go.loadValue(sp + 8)
            m = getattr(v, self.go.loadString(sp + 16))
            args = self.go.loadSliceOfValues(sp + 32)
            result = m(*args)
            sp = self.go._inst["getsp"](self.go.wasm_store) >> 0
            self.go.storeValue(sp + 56, result)
            self.mem.setUint8(sp + 64, 1)
        except Exception as e:
            traceback.print_exc()
            sp = self.go._inst["getsp"](self.go.wasm_store) >> 0
            self.go.storeValue(sp + 56, e)
            self.mem.setUint8(sp + 64, 0)

    def sysjs_valueInvoke(self, *args):
        pass

    def sysjs_valueNew(self, *args):
        pass

    def sysjs_valueLength(self, sp):
        sp = (sp >> 0) & 0xFFFFFFFF
        self.go.setInt64(sp + 16, len(self.go.loadValue(sp + 8)))

    def sysjs_valuePrepareString(self, sp):
        sp = (sp >> 0) & 0xFFFFFFFF
        data = self.go.loadValue(sp + 8)
        if type(data) is float:
            data = int(data)
        str_data = str(data).encode('utf-8')
        self.go.storeValue(sp + 16, str_data)
        self.go.setInt64(sp + 24, len(str_data))

    def sysjs_valueLoadString(self, sp):
        sp = (sp >> 0) & 0xFFFFFFFF
        str_data = self.go.loadValue(sp + 8)
        byte_aray, start, length = self.go.loadSlice(sp + 16)
        self.mem.write(str_data, start)

    def sysjs_valueInstanceOf(self, *args):
        pass

    def sysjs_copyBytesToGo(self, *args):
        pass

    def sysjs_copyBytesToJS(self, *args):
        pass

    def debug(self, *args):
        pass


def wasm_encrypt(wasm_files, json_data):
    if getattr(global_this, 'bder', None) is not None:
        return global_this.bder(json.dumps(json_data), "0")
    engine = wasmtime.Engine()
    modun = wasmtime.Module(engine, wasm_files)
    store = wasmtime.Store(engine)
    go_obj = GO(store)
    instance = wasmtime.Instance(store, modun, imports=go_obj.importObject(modun.imports))
    run_as_lib = instance.exports(store)
    go_obj.run(run_as_lib)
    return global_this.bder(json.dumps(json_data), "0")


import dataclasses
import json
import struct
import wasmtime


class Memory:
    def __init__(self, store_wasm, mem: wasmtime.Memory):
        self.store_wasm = store_wasm
        self.mem = mem
        self.wasm_page_size = 65536

    def read(self, start_address, end_address):
        return self.mem.read(self.store_wasm, start_address, end_address)

    def write(self, value, start_address):
        need = start_address + len(value)
        if self.mem.size(self.store_wasm) * self.wasm_page_size < need:
            print("need", need, "size", self.mem.size(self.store_wasm) * self.wasm_page_size)
            if self.wasm_page_size > need:
                grow_page = 1
            else:
                grow_page = int(need / self.wasm_page_size)
                if grow_page > int(grow_page):
                    grow_page += 1
            self.mem.grow(self.store_wasm, int(grow_page))
        return self.mem.write(self.store_wasm, value, start_address)

    def getBigInt64(self, address, littleEndian=False):
        data = self.read(address, address + 8)
        if littleEndian:
            value, = struct.unpack('<q', data)
        else:
            value, = struct.unpack('>q', data)
        return value

    def getBigUint64(self, address, littleEndian=False):
        data = self.read(address, address + 8)
        if littleEndian:
            value, = struct.unpack('<Q', data)
        else:
            value, = struct.unpack('>Q', data)
        return value

    def getFloat16(self, address, littleEndian=False):
        data = self.read(address, address + 2)
        value, = struct.unpack('<e', data)  # Replace '<e' with the correct format
        return value

    def getFloat32(self, address, littleEndian=False):
        data = self.read(address, address + 4)
        if littleEndian:
            value, = struct.unpack('<f', data)
        else:
            value, = struct.unpack('>f', data)
        return value

    def getFloat64(self, address, littleEndian=False):
        data = self.read(address, address + 8)
        if littleEndian:
            value, = struct.unpack('<d', data)
        else:
            value, = struct.unpack('>d', data)
        return value

    def getInt16(self, address, littleEndian=False):
        data = self.read(address, address + 2)
        if littleEndian:
            value, = struct.unpack('<h', data)
        else:
            value, = struct.unpack('>h', data)
        return value

    def getInt32(self, address, littleEndian=False):
        data = self.read(address, address + 4)
        if littleEndian:
            value, = struct.unpack('<i', data)
        else:
            value, = struct.unpack('>i', data)
        return value

    def getInt8(self, address, littleEndian=False):
        data = self.read(address, address + 1)
        value, = struct.unpack('b', data)
        return value

    def getUint16(self, address, littleEndian=False):
        data = self.read(address, address + 2)
        if littleEndian:
            value, = struct.unpack('<H', data)
        else:
            value, = struct.unpack('>H', data)
        return value

    def getUint32(self, address, littleEndian=False):
        data = self.read(address, address + 4)
        if littleEndian:
            value, = struct.unpack('<I', data)
        else:
            value, = struct.unpack('>I', data)
        return value

    def getUint8(self, address, littleEndian=False):
        data = self.read(address, address + 1)
        value, = struct.unpack('B', data)
        return value

    def setBigInt64(self, address, value, littleEndian=False):
        format_str = '<q' if littleEndian else '>q'
        data = struct.pack(format_str, value)
        self.write(data, address)

    def setBigUint64(self, address, value, littleEndian=False):
        format_str = '<Q' if littleEndian else '>Q'
        data = struct.pack(format_str, value)
        self.write(data, address)

    def setFloat16(self, address, value, littleEndian=False):
        format_str = '<e'  # Replace '<e' with the correct format
        data = struct.pack(format_str, value)
        self.write(data, address)

    def setFloat32(self, address, value, littleEndian=False):
        format_str = '<f' if littleEndian else '>f'
        data = struct.pack(format_str, value)
        self.write(data, address)

    def setFloat64(self, address, value, littleEndian=False):
        format_str = '<d' if littleEndian else '>d'
        data = struct.pack(format_str, value)
        self.write(data, address)

    def setInt16(self, address, value, littleEndian=False):
        format_str = '<h' if littleEndian else '>h'
        data = struct.pack(format_str, value)
        self.write(data, address)

    def setInt32(self, address, value, littleEndian=False):
        format_str = '<i' if littleEndian else '>i'
        data = struct.pack(format_str, value)
        self.write(data, address)

    def setInt8(self, address, value, littleEndian=False):
        data = struct.pack('b', value)
        self.write(data, address)

    def setUint16(self, address, value, littleEndian=False):
        format_str = '<H' if littleEndian else '>H'
        data = struct.pack(format_str, value)
        self.write(data, address)

    def setUint32(self, address, value, littleEndian=False):
        value = int(value)
        value &= 0xFFFFFFFF
        format_str = '<I' if littleEndian else '>I'
        data = struct.pack(format_str, value)
        self.write(data, address)

    def setUint8(self, address, value, littleEndian=False):
        data = struct.pack('B', value)
        self.write(data, address)


class fs_object:
    def __init__(self):
        self.outputBuf = ""

        @dataclasses.dataclass(unsafe_hash=True)
        class constants_object:
            O_WRONLY: int
            O_RDWR: int
            O_CREAT: int
            O_TRUNC: int
            O_APPEND: int
            O_EXCL: int

        self.constants = constants_object(
            **{"O_WRONLY": -1, "O_RDWR": -1, "O_CREAT": -1, "O_TRUNC": -1, "O_APPEND": -1, "O_EXCL": -1})

    def enosys(self):
        raise NotImplementedError

    def writeSync(self, fd, buf: bytes):
        self.outputBuf += buf.decode()
        nl = self.outputBuf.rfind("\n")
        if nl != -1:
            print(self.outputBuf[:nl])
            self.outputBuf = self.outputBuf[nl + 1:]
        return len(buf)

    def write(self, fd, buf, offset, length, position, callback):
        n = self.writeSync(fd, buf)
        callback(None, n)

class process_object:
    def __init__(self):
        self.ppid = -1
        self.pid = -1
        pass

    def getuid(self):
        return -1

    def getgid(self):
        return -1

    def geteuid(self):
        return -1

    def getegid(self):
        return -1


class dict_warper:
    def __init__(self, dict_data):
        for key, value in dict_data.items():
            setattr(self, key, value)

    def to_dict(self):
        return {key: getattr(self, key) for key in dir(self) if not (key.startswith("__") or key == "to_dict")}


class hash_list(list):
    def __hash__(self):
        return hash(json.dumps(self))

    def get(self, index, default=None):
        try:
            return self[index]
        except IndexError:
            return default



import asyncio
import concurrent.futures
import datetime
import base64
import hashlib
import typing
import aiohttp
from .capcha_ocr import CapchaProcessing, CapchaOCR
from .main import MBBankError, MBBank
from .wasm_helper import wasm_encrypt
from .main import headers_default

pool = concurrent.futures.ThreadPoolExecutor()


def get_now_time():
    now = datetime.datetime.now()
    microsecond = int(now.strftime("%f")[:2])
    return now.strftime(f"%Y%m%d%H%M{microsecond}")


class MBBankAsync(MBBank):
    """Core Async class

    Attributes:
        deviceIdCommon (str): Device id common
        sessionId (str or None): Current Session id

    Args:
        username (str): MBBank Account Username
        password (str): MBBank Account Password
        proxy (str, optional): Proxy url. Example: "http://127.0.0.1:8080". Defaults to None.
        ocr_class (CapchaProcessing, optional): CapchaProcessing class. Defaults to TesseractOCR().
    """

    def __init__(self, *, username: str, password: str, proxy: dict=None, ocr_class=None):
        super().__init__(username=username, password=password, proxy=proxy, ocr_class=ocr_class)
        # convert proxy dict by requests to aiohttp format
        if len(self.proxy.values()):
            self.proxy = self.proxy.values()[0]
        else:
            self.proxy = None

    async def _get_wasm_file(self):
        if self._wasm_cache is not None:
            return self._wasm_cache
        async with aiohttp.ClientSession() as s:
            async with s.get("https://online.mbbank.com.vn/assets/wasm/main.wasm", headers=headers_default,
                             proxy=self.proxy) as r:
                self._wasm_cache = await r.read()
        return self._wasm_cache

    async def _authenticate(self):
        while True:
            self._userinfo = None
            self.sessionId = None
            self._temp = {}
            rid = f"{self._userid}-{get_now_time()}"
            json_data = {
                'sessionId': "",
                'refNo': rid,
                'deviceIdCommon': self.deviceIdCommon,
            }
            headers = headers_default.copy()
            headers["X-Request-Id"] = rid
            headers["Deviceid"] = self.deviceIdCommon
            headers["Refno"] = rid
            async with aiohttp.ClientSession() as s:
                async with s.post("https://online.mbbank.com.vn/retail-web-internetbankingms/getCaptchaImage",
                                  headers=headers, json=json_data, proxy=self.proxy) as r:
                    data_out = await r.json()
            img_bytes = base64.b64decode(data_out["imageString"])
            text = await asyncio.get_event_loop().run_in_executor(
                pool, self.ocr_class.process_image, img_bytes
            )
            payload = {
                "userId": self._userid,
                "password": hashlib.md5(self._password.encode()).hexdigest(),
                "captcha": text,
                'sessionId': "",
                'refNo': f'{self._userid}-{get_now_time()}',
                'deviceIdCommon': self.deviceIdCommon,
                "ibAuthen2faString": self.FPR,
            }
            wasm_bytes = await self._get_wasm_file()
            loop = asyncio.get_running_loop()
            dataEnc = await loop.run_in_executor(pool, wasm_encrypt, wasm_bytes, payload)
            async with aiohttp.ClientSession() as s:
                async with s.post("https://online.mbbank.com.vn/retail_web/internetbanking/doLogin",
                                  headers=headers_default, json={"dataEnc": dataEnc}, proxy=self.proxy) as r:
                    data_out = await r.json()
            if data_out["result"]["ok"]:
                self.sessionId = data_out["sessionId"]
                self._userinfo = data_out
                return
            elif data_out["result"]["responseCode"] == "GW283":
                pass
            else:
                err_out = data_out["result"]
                raise Exception(f"{err_out['responseCode']} | {err_out['message']}")

    async def _req(self, url, *, json=None, headers=None):
        if headers is None:
            headers = {}
        if json is None:
            json = {}
        while True:
            if self.sessionId is None:
                await self._authenticate()
            rid = f"{self._userid}-{get_now_time()}"
            json_data = {
                'sessionId': self.sessionId if self.sessionId is not None else "",
                'refNo': rid,
                'deviceIdCommon': self.deviceIdCommon,
            }
            json_data.update(json)
            headers.update(headers_default)
            headers["X-Request-Id"] = rid
            headers["RefNo"] = rid
            headers["DeviceId"] = self.deviceIdCommon
            async with aiohttp.ClientSession() as s:
                async with s.post(url, headers=headers, json=json_data, proxy=self.proxy) as r:
                    data_out = await r.json()
            if data_out["result"] is None:
                await self.getBalance()
            elif data_out["result"]["ok"]:
                data_out.pop("result", None)
                break
            elif data_out["result"]["responseCode"] == "GW200":
                await self._authenticate()
            else:
                err_out = data_out["result"]
                raise MBBankError(err_out)
        return data_out

    async def getTransactionAccountHistory(self, *, accountNo: str = None, from_date: datetime.datetime,
                                           to_date: datetime.datetime):
        """
        Get account transaction history

        Args:
            accountNo (str, optional): Sub account number Defaults to Main Account number.
            from_date (datetime.datetime): transaction from date
            to_date (datetime.datetime): transaction to date

        Returns:
            success (dict): account transaction history

        Raises:
            MBBankError: if api response not ok
        """
        json_data = {
            'accountNo': self._userid if accountNo is None else accountNo,
            'fromDate': from_date.strftime("%d/%m/%Y"),
            'toDate': to_date.strftime("%d/%m/%Y"),  # max 3 months
        }
        data_out = await self._req(
            "https://online.mbbank.com.vn/api/retail-transactionms/transactionms/get-account-transaction-history",
            json=json_data)
        return data_out

    async def getBalance(self):
        """
        Get all main account and subaccount balance

        Returns:
            success (dict): list account balance

        Raises:
            MBBankError: if api response not ok
        """
        data_out = await self._req("https://online.mbbank.com.vn/api/retail-web-accountms/getBalance")
        return data_out

    async def getBalanceLoyalty(self):
        """
        Get Account loyalty rank and Member loyalty point

        Returns:
            success (dict): loyalty point

        Raises:
            MBBankError: if api response not ok
        """
        data_out = await self._req("https://online.mbbank.com.vn/api/retail_web/loyalty/getBalanceLoyalty")
        return data_out

    async def getInterestRate(self, currency: str = "VND"):
        """
        Get saving interest rate

        Args:
            currency (str, optional): currency ISO 4217 format. Defaults to "VND" (Vietnam Dong).

        Returns:
            success (dict): interest rate

        Raises:
            MBBankError: if api response not ok
        """
        json_data = {
            "productCode": "TIENGUI.KHN.EMB",
            "currency": currency
        }
        data_out = await self._req("https://online.mbbank.com.vn/api/retail_web/saving/getInterestRate", json=json_data)
        return data_out

    async def getFavorBeneficiaryList(self, *, transactionType: typing.Literal["TRANSFER", "PAYMENT"],
                                      searchType: typing.Literal["MOST", "LATEST"]):
        """
        Get all favor or most transfer beneficiary list from your account

        Args:
            transactionType (Literal["TRANSFER", "PAYMENT"]): transaction type
            searchType (Literal["MOST", "LATEST"]): search type

        Returns:
            success (dict): favor beneficiary list

        Raises:
            MBBankError: if api response not ok
        """
        json_data = {
            "transactionType": transactionType,
            "searchType": searchType
        }
        data_out = await self._req(
            "https://online.mbbank.com.vn/api/retail_web/internetbanking/getFavorBeneficiaryList", json=json_data)
        return data_out

    async def getCardList(self):
        """
        Get all card list from your account

        Returns:
            success (dict): card list

        Raises:
            MBBankError: if api response not ok
        """
        data_out = await self._req("https://online.mbbank.com.vn/api/retail_web/card/getList")
        return data_out

    async def getSavingList(self):
        """
        Get all saving list from your account

        Returns:
            success (dict): saving list

        Raises:
            MBBankError: if api response not ok
        """
        data_out = await self._req("https://online.mbbank.com.vn/api/retail_web/saving/getList")
        return data_out

    async def getSavingDetail(self, accNo: str, accType: typing.Literal["OSA", "SBA"]):
        """
        Get saving detail by account number

        Args:
            accNo (str): saving account number
            accType (Literal["OSA", "SBA"]): saving account type | OSA: Online Saving Account, SBA: Saving Bank Account

        Returns:
            success (dict): saving detail

        Raises:
            MBBankError: if api response not ok
        """
        json_data = {
            "accNo": accNo,
            "accType": accType
        }
        data_out = await self._req("https://online.mbbank.com.vn/api/retail_web/saving/getDetail", json=json_data)
        return data_out


    async def getLoanList(self):
        """
        Get all loan list from your account

        Returns:
            success (dict): loan list

        Raises:
            MBBankError: if api response not ok
        """
        data_out = await self._req("https://online.mbbank.com.vn/api/retail-web-onlineloanms/loan/getList")
        return data_out

    async def getCardTransactionHistory(self, cardNo: str, from_date: datetime.datetime, to_date: datetime.datetime):
        """
        Get card transaction history

        Args:
            cardNo (str): card number get from getCardList
            from_date (datetime.datetime): from date
            to_date (datetime.datetime): to date

        Returns:
            success (dict): card transaction history

        Raises:
            MBBankError: if api response not ok
        """
        json_data = {
            "accountNo": cardNo,
            "fromDate": from_date.strftime("%d/%m/%Y"),
            "toDate": to_date.strftime("%d/%m/%Y"),  # max 3 months
            "historyNumber": "",
            "historyType": "DATE_RANGE",
            "type": "CARD",
        }
        data_out = await self._req("https://online.mbbank.com.vn/api/retail_web/common/getTransactionHistory",
                                   json=json_data)
        return data_out

    async def getBankList(self):
        """
        Get transfer all bank list

        Returns:
            success (dict): bank list

        Raises:
            MBBankError: if api response not ok
        """
        data_out = await self._req("https://online.mbbank.com.vn/api/retail_web/common/getBankList")
        return data_out

    async def getAccountByPhone(self, phone: str):
        """
        Get transfer account info by phone (MBank internal account only)

        Args:
            phone (str): MBBank account phone number

        Returns:
            success (dict): account info

        """
        json_data = {
            "phone": phone
        }
        data_out = await self._req("https://online.mbbank.com.vn/api/retail_web/common/getAccountByPhone",
                                   json=json_data)
        return data_out

    async def userinfo(self):
        """
        Get current user info

        Returns:
            success (dict): user info

        Raises:
            MBBankError: if api response not ok
        """
        if self._userinfo is None:
            await self._authenticate()
        else:
            await self.getBalance()
        return self._userinfo



import datetime
import base64
import hashlib
import typing
import requests
from .capcha_ocr import CapchaOCR, CapchaProcessing
from .wasm_helper import wasm_encrypt

headers_default = {
    'Cache-Control': 'max-age=0',
    'Accept': 'application/json, text/plain, */*',
    'Authorization': 'Basic RU1CUkVUQUlMV0VCOlNEMjM0ZGZnMzQlI0BGR0AzNHNmc2RmNDU4NDNm',
    'User-Agent': "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36",
    "Origin": "https://online.mbbank.com.vn",
    "Referer": "https://online.mbbank.com.vn/pl/login?returnUrl=%2F",
    "App": "MB_WEB",
    "Sec-Ch-Ua": '"Not.A/Brand";v="8", "Chromium";v="134", "Google Chrome";v="134"',
    "Sec-Ch-Ua-Mobile": "?0",
    "Sec-Ch-Ua-Platform": '"Windows"',
    "Sec-Fetch-Dest": "empty",
    "Sec-Fetch-Mode": "cors",
    "Sec-Fetch-Site": "same-origin",
}


class MBBankError(Exception):
    def __init__(self, err_out):
        self.code = err_out['responseCode']
        self.message = err_out['message']
        super().__init__(f"{err_out['responseCode']} | {err_out['message']}")


class MBBank:
    """Core class

    Attributes:
        deviceIdCommon (str): Device id common
        sessionId (str or None): Current Session id

    Args:
        username (str): MBBank Account Username
        password (str): MBBank Account Password
        proxy (str, optional): Proxy url. Example: "http://127.0.0.1:8080". Defaults to None.
        ocr_class (CapchaProcessing, optional): instance of CapchaProcessing class. Defaults to CapchaOCR().
    """

    FPR = "c7a1beebb9400375bb187daa33de9659"

    def __init__(self, *, username: str, password: str, proxy: dict=None, ocr_class=None):
        self._userid = username
        self._password = password
        self._wasm_cache = None
        if proxy is not None:
            proxy_protocol = proxy.split("://")[0]
            self.proxy = {proxy_protocol: proxy}
        else:
            self.proxy = {}
        self.ocr_class = CapchaOCR()
        if ocr_class is not None:
            if not isinstance(ocr_class, CapchaProcessing):
                raise ValueError("ocr_class must be instance of CapchaProcessing")
            self.ocr_class = ocr_class
        self.sessionId = None
        self._userinfo = None
        self._temp = {}
        self.deviceIdCommon = f'i1vzyjp5-mbib-0000-0000-{self._get_now_time()}'

    def _get_now_time(self):
        now = datetime.datetime.now()
        microsecond = int(now.strftime("%f")[:2])
        return now.strftime(f"%Y%m%d%H%M{microsecond}")

    def _req(self, url, *, json=None, headers=None):
        if headers is None:
            headers = {}
        if json is None:
            json = {}
        while True:
            if self.sessionId is None:
                self._authenticate()
            rid = f"{self._userid}-{self._get_now_time()}"
            json_data = {
                'sessionId': self.sessionId if self.sessionId is not None else "",
                'refNo': rid,
                'deviceIdCommon': self.deviceIdCommon,
            }
            json_data.update(json)
            headers.update(headers_default)
            headers["X-Request-Id"] = rid
            headers["RefNo"] = rid
            headers["DeviceId"] = self.deviceIdCommon
            with requests.Session() as s:
                with s.post(url, headers=headers, json=json_data,
                            proxies=self.proxy) as r:
                    data_out = r.json()
            if data_out["result"] is None:
                self.getBalance()
            elif data_out["result"]["ok"]:
                data_out.pop("result", None)
                break
            elif data_out["result"]["responseCode"] == "GW200":
                self._authenticate()
            else:
                err_out = data_out["result"]
                raise MBBankError(err_out)
        return data_out

    def _get_wasm_file(self):
        if self._wasm_cache is not None:
            return self._wasm_cache
        file_data = requests.get("https://online.mbbank.com.vn/assets/wasm/main.wasm",
                                 proxies=self.proxy).content
        self._wasm_cache = file_data
        return file_data

    def _authenticate(self):
        while True:
            self._userinfo = None
            self.sessionId = None
            self._temp = {}
            rid = f"{self._userid}-{self._get_now_time()}"
            json_data = {
                'sessionId': "",
                'refNo': rid,
                'deviceIdCommon': self.deviceIdCommon,
            }
            headers = headers_default.copy()
            headers["X-Request-Id"] = rid
            headers["Deviceid"] = self.deviceIdCommon
            headers["Refno"] = rid
            with requests.Session() as s:
                with s.post("https://online.mbbank.com.vn/retail-web-internetbankingms/getCaptchaImage",
                            headers=headers, json=json_data,
                            proxies=self.proxy) as r:
                    data_out = r.json()
            img_bytes = base64.b64decode(data_out["imageString"])
            text = self.ocr_class.process_image(img_bytes)
            payload = {
                "userId": self._userid,
                "password": hashlib.md5(self._password.encode()).hexdigest(),
                "captcha": text,
                'sessionId': "",
                'refNo': f'{self._userid}-{self._get_now_time()}',
                'deviceIdCommon': self.deviceIdCommon,
                "ibAuthen2faString": self.FPR,
            }
            wasm_bytes = self._get_wasm_file()
            dataEnc = wasm_encrypt(wasm_bytes, payload)
            with requests.Session() as s:
                with s.post("https://online.mbbank.com.vn/retail_web/internetbanking/doLogin",
                            headers=headers_default, json={"dataEnc": dataEnc},
                            proxies=self.proxy) as r:
                    data_out = r.json()
            if data_out["result"]["ok"]:
                self.sessionId = data_out["sessionId"]
                self._userinfo = data_out
                return
            elif data_out["result"]["responseCode"] == "GW283":
                pass
            else:
                err_out = data_out["result"]
                raise Exception(f"{err_out['responseCode']} | {err_out['message']}")

    def getTransactionAccountHistory(self, *, accountNo: str = None, from_date: datetime.datetime,
                                     to_date: datetime.datetime):
        """
        Get account transaction history

        Args:
            accountNo (str, optional): Sub account number Defaults to Main Account number.
            from_date (datetime.datetime): transaction from date
            to_date (datetime.datetime): transaction to date

        Returns:
            success (dict): account transaction history

        Raises:
            MBBankError: if api response not ok
        """
        if self._userinfo is None:
            self._authenticate()
        json_data = {
            'accountNo': self._userid if accountNo is None else accountNo,
            'fromDate': from_date.strftime("%d/%m/%Y"),
            'toDate': to_date.strftime("%d/%m/%Y"),  # max 3 months
        }
        data_out = self._req(
            "https://online.mbbank.com.vn/api/retail-transactionms/transactionms/get-account-transaction-history",
            json=json_data)
        return data_out

    def getBalance(self):
        """
        Get all main account and subaccount balance

        Returns:
            success (dict): list account balance

        Raises:
            MBBankError: if api response not ok
        """
        if self._userinfo is None:
            self._authenticate()
        data_out = self._req("https://online.mbbank.com.vn/api/retail-web-accountms/getBalance")
        return data_out

    def getBalanceLoyalty(self):
        """
        Get Account loyalty rank and Member loyalty point

        Returns:
            success (dict): loyalty point

        Raises:
            MBBankError: if api response not ok
        """
        data_out = self._req("https://online.mbbank.com.vn/api/retail_web/loyalty/getBalanceLoyalty")
        return data_out

    def getInterestRate(self, currency: str = "VND"):
        """
        Get saving interest rate

        Args:
            currency (str, optional): currency ISO 4217 format. Defaults to "VND" (Vietnam Dong).

        Returns:
            success (dict): interest rate

        Raises:
            MBBankError: if api response not ok
        """
        json_data = {
            "productCode": "TIENGUI.KHN.EMB",
            "currency": currency,
        }
        data_out = self._req("https://online.mbbank.com.vn/api/retail_web/saving/getInterestRate", json=json_data)
        return data_out

    def getFavorBeneficiaryList(self, *, transactionType: typing.Literal["TRANSFER", "PAYMENT"],
                                searchType: typing.Literal["MOST", "LATEST"]):
        """
        Get all favor or most transfer beneficiary list from your account

        Args:
            transactionType (Literal["TRANSFER", "PAYMENT"]): transaction type
            searchType (Literal["MOST", "LATEST"]): search type

        Returns:
            success (dict): favor beneficiary list

        Raises:
            MBBankError: if api response not ok
        """
        json_data = {
            "transactionType": transactionType,
            "searchType": searchType
        }
        data_out = self._req(
            "https://online.mbbank.com.vn/api/retail_web/internetbanking/getFavorBeneficiaryList", json=json_data)
        return data_out

    def getCardList(self):
        """
        Get all card list from your account

        Returns:
            success (dict): card list

        Raises:
            MBBankError: if api response not ok
        """
        data_out = self._req("https://online.mbbank.com.vn/api/retail_web/card/getList")
        return data_out

    def getSavingList(self):
        """
        Get all saving list from your account

        Returns:
            success (dict): saving list

        Raises:
            MBBankError: if api response not ok
        """
        data_out = self._req("https://online.mbbank.com.vn/api/retail_web/saving/getList")
        return data_out

    def getSavingDetail(self, accNo: str, accType: typing.Literal["OSA", "SBA"]):
        """
        Get saving detail by account number

        Args:
            accNo (str): saving account number
            accType (Literal["OSA", "SBA"]): saving account type | OSA: Online Saving Account, SBA: Saving Bank Account

        Returns:
            success (dict): saving detail

        Raises:
            MBBankError: if api response not ok
        """
        json_data = {
            "accNo": accNo,
            "accType": accType
        }
        data_out = self._req("https://online.mbbank.com.vn/api/retail_web/saving/getDetail", json=json_data)
        return data_out

    def getLoanList(self):
        """
        Get all loan list from your account

        Returns:
            success (dict): loan list

        Raises:
            MBBankError: if api response not ok
        """
        data_out = self._req("https://online.mbbank.com.vn/api/retail-web-onlineloanms/loan/getList")
        return data_out

    def getCardTransactionHistory(self, cardNo: str, from_date: datetime.datetime, to_date: datetime.datetime):
        """
        Get card transaction history

        Args:
            cardNo (str): card number get from getCardList
            from_date (datetime.datetime): from date
            to_date (datetime.datetime): to date

        Returns:
            success (dict): card transaction history

        Raises:
            MBBankError: if api response not ok
        """
        json_data = {
            "accountNo": cardNo,
            "fromDate": from_date.strftime("%d/%m/%Y"),
            "toDate": to_date.strftime("%d/%m/%Y"),  # max 3 months
            "historyNumber": "",
            "historyType": "DATE_RANGE",
            "type": "CARD",
        }
        data_out = self._req("https://online.mbbank.com.vn/api/retail_web/common/getTransactionHistory", json=json_data)
        return data_out

    def getBankList(self):
        """
        Get transfer all bank list

        Returns:
            success (dict): bank list

        Raises:
            MBBankError: if api response not ok
        """
        data_out = self._req("https://online.mbbank.com.vn/api/retail_web/common/getBankList")
        return data_out

    def getAccountByPhone(self, phone: str):
        """
        Get transfer account info by phone (MBank internal account only)

        Args:
            phone (str): MBBank account phone number

        Returns:
            success (dict): account info

        """
        json_data = {
            "phone": phone
        }
        data_out = self._req("https://online.mbbank.com.vn/api/retail_web/common/getAccountByPhone", json=json_data)
        return data_out

    def userinfo(self):
        """
        Get current user info

        Returns:
            success (dict): user info

        Raises:
            MBBankError: if api response not ok
        """
        if self._userinfo is None:
            self._authenticate()
        else:
            self.getBalance()
        return self._userinfo
    def inquiryAccountName(self, *, creditAccount: str, bankCode: str, 
                      debitAccount: str = None, creditAccountType: str = "ACCOUNT", 
                      transfer_type: str = "FAST", remark: str = ""):
        """
        Get account name inquiry for transfer

        Args:
            creditAccount (str): Recipient account number
            bankCode (str): Bank code of recipient bank
            debitAccount (str, optional): Sender account number. Defaults to main account.
            creditAccountType (str, optional): Credit account type. Defaults to "ACCOUNT".
            transfer_type (str, optional): Transfer type. Defaults to "FAST".
            remark (str, optional): Transfer remark. Defaults to "".

        Returns:
            success (dict): account name inquiry result

        Raises:
            MBBankError: if api response not ok
        """
        if self._userinfo is None:
            self._authenticate()
        
        json_data = {
            "creditAccount": creditAccount,
            "creditAccountType": creditAccountType,
            "bankCode": bankCode,
            "debitAccount": self._userid if debitAccount is None else debitAccount,
            "type": transfer_type,
            "remark": remark,
        }
        
        data_out = self._req("https://online.mbbank.com.vn/api/retail_web/transfer/inquiryAccountName", 
                            json=json_data)
        return data_out



import re
from setuptools import setup

with open("requirements.txt", "r") as f:
    req = f.read().splitlines()

with open("README.MD", "r") as f:
    ldr = f.read()

with open('mbbankchecker/__init__.py') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

setup(
    name='MBbankchecker',
    version=version,
    description='bankchecker base on MBBank of The DT',
    author='JussKynn',
    packages=["mbbankchecker", "mbbankchecker.wasm_helper"],
    install_requires=req,
    include_package_data=True
)
